# ========================================================================= 
# THIS CODE-SAMPLE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER  
# EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES  
# OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE. 
# 
# This sample is not supported under any Microsoft standard support program  
# or service. The code sample is provided AS IS without warranty of any kind.  
# Microsoft further disclaims all implied warranties including, without  
# limitation, any implied warranties of merchantability or of fitness for a  
# particular purpose. The entire risk arising out of the use or performance 
# of the sample and documentation remains with you. In no event shall  
# Microsoft, its authors, or anyone else involved in the creation,  
# production, or delivery of the script be liable for any damages whatsoever  
# (including, without limitation, damages for loss of business profits,  
# business interruption, loss of business information, or other pecuniary  
# loss) arising out of  the use of or inability to use the sample or  
# documentation, even if Microsoft has been advised of the possibility of  
# such damages. 
#=========================================================================  
 
 
#region ============================ Synopsis ============================ 
# When performing adprep /domainprep from Windows Server 2016 there will 
# be an unwanted AccessControlEntry (ACE) in the SecurityDescriptor of the  
# domain naming context for the Enterprise Key Admins group  
# ( SID = <domain sid of forest root domain>-527 ) that grants 
# full access to the naming context and all its cildren. 
# The desired ACE should only grant ReadProperty / WritePropety to 
# msDS-KeyCredentialLink attribute. 
# This sample code  
# - inspects the SecurityDescriptor of the domain naming context 
# - removes ACEs for Enterprise Key Admins group that are not of the 
#   desired type 
# - adds the desired ACE for Enterprise Key Admins group 
#endregion =============================================================== 
 
 
#region ========================= Prerequisites ========================== 
#      Code should be run as member of Enterprise Admins group 
#endregion =============================================================== 
 
 
#region ============================= Usage ============================== 
# Sample code takes -ForestName <FQDN of target forest> as call parameter 
# and -Force as switch. 
#  
#  If no forest name is passed the forest of the calling user is targeted 
#  otherwise the given forest name is used. 
# 
#  If -Force is passed the default 527 ACE is added (if not yet present) 
#  even though there was no FullControl 527 ACE. 
#endregion =============================================================== 
 
#region params 
 
param 
( 
    [Parameter(Position=0,Mandatory=$false)] 
    [string]$ForestName = [String]::Empty, 
 
    [Switch]$Force 
) 
 
#endregion 
 
 
#region functions 
 
 
#region retrieve forest info such as 
# - forest name 
# - SID of forest root domain 
# - list of domains in forest 
#endregion 
function GetForest([string]$forestname) 
{     
    $ret = @{ Forestname = $forestname ; RootSid = [String]::Empty ; Domains = $null ; Success = $true } 
 
    [System.DirectoryServices.ActiveDirectory.Forest] $forest 
 
    try 
    { 
        # no forest name given -> use current forest 
        if ($forestname -eq [String]::Empty) 
        {  
            Write-Host "Trying to retrieve forest info from current forest" 
 
            [System.DirectoryServices.ActiveDirectory.Forest]$forest = [System.DirectoryServices.ActiveDirectory.Forest]::GetCurrentForest() 
        } 
 
        # collect data from given forest 
        else 
        { 
            Write-Host "Trying to retrieve forest info from current $forestname" 
 
            [System.DirectoryServices.ActiveDirectory.DirectoryContextType] $ctxType = [System.DirectoryServices.ActiveDirectory.DirectoryContextType]::Forest 
             
            [System.DirectoryServices.ActiveDirectory.DirectoryContext]$ctx = New-Object System.DirectoryServices.ActiveDirectory.DirectoryContext($ctxType, $forestname) 
             
            [System.DirectoryServices.ActiveDirectory.Forest] $forest = [System.DirectoryServices.ActiveDirectory.Forest]::GetForest($ctx) 
                         
        } 
             
        if ($forest.RootDomain -eq $null) 
        { 
            $ret.Success = $false 
 
            Write-Host "ERROR: Failed to retrieve forest info from $forestName" 
 
            return $ret 
 
        } 
 
        Write-Host "Detected forest $($forest.Name)" 
         
        # get forest root domain SID 
        Write-Host "Trying to retrieve forest root domain SID" 
 
        [byte[]] $bsid = $forest.RootDomain.GetDirectoryEntry().Properties["objectSid"].Value 
 
        [System.Security.Principal.SecurityIdentifier] $sid = New-Object System.Security.Principal.SecurityIdentifier($bsid, 0) 
 
        Write-Host "Forest root domain SID: $($sid.Value)" 
 
        # build forest info data 
        # the forest root domain SID 
        $ret.RootSid = $sid.Value 
 
        # the forest name 
        $ret.ForestName = $forest.Name 
 
        # the list of domains in forest 
        $ret.Domains = $forest.Domains 
    } 
 
    catch 
    { 
        Write-Host "ERROR: Failed to retrieve forest info: $($_.Exception.Message)" 
 
        # we failed -> no need to proceed in main processing 
        $ret.Success = $false 
    } 
 
    return $ret 
} 
 
#region inspect domain namingContext SecurityDescriptor 
# and handle appropriately. 
#endregion 
function HandleDomain([System.DirectoryServices.ActiveDirectory.Domain] $domain, [bool]$force) 
{     
    Write-Host "`r`nInspecting domain $($domain.Name)" 
 
    try 
    { 
        # get S.DS.DirectoryEntry of current domain 
        [System.DirectoryServices.DirectoryEntry] $dedom = $domain.GetDirectoryEntry() 
         
        # get SecurityDescriptor of domain nc 
        $sd = $dedom.ObjectSecurity 
 
        # inspect SecurityDescriptor and edit appropriately 
        $res = HandleSD $sd $force 
 
        # did we edit the SecurityDescriptor? 
        if ($res.MustCommit -eq $true) 
        { 
            Write-Host "`r`nCommiting SD changes"             
 
            # write back SecurityDescriptor to S.DS.DirectoryEntry of current domain 
            $dedom.ObjectSecurity = [System.DirectoryServices.ActiveDirectorySecurity]($res.NewSD) 
 
            # commit changes from ADSI Pproperty Cache to Active Directory 
            $dedom.CommitChanges() 
 
            Write-Host "Commiting SD successfull" 
        } 
 
        # SecurityDescriptor as expcected 
        else 
        { 
            Write-Host "`r`nNothing to commit" 
        } 
    } 
 
    catch 
    { 
        Write-Host "ERROR: HandleDomain: $($_.Exception.Message)" 
    } 
} 
 
#region inspect SecurityDescriptor and edit appropriately. 
# If we have an AccessControlEntry where trustee matches EnterpriseKeyAdmins group 
# we check for correct AccessControlEntry settings. 
# If no match -> remove theis AccessControlEntry. 
# If correct AccessControlEntry is present as well -> no action necessary. 
# Otherwise add desired default AccessControlEntry for EnterpriseKeyAdmins group 
#endregion 
function HandleSD([System.DirectoryServices.ActiveDirectorySecurity] $sd, [bool]$force) 
{ 
    $ret = @{ MustCommit = $false ; NewSD = $null } 
 
    # SecurityDescriptor to be returned 
    [System.DirectoryServices.ActiveDirectorySecurity] $retsd = $sd 
 
    # get DiscretionaryAccessControlList from SecurityDescriptor 
    [System.Security.AccessControl.AuthorizationRuleCollection] $acl = $retsd.GetAccessRules($true, $false, [System.Security.Principal.SecurityIdentifier]) 
 
    [bool] $match = $false 
 
    [bool] $mustedit = $false 
 
    [System.DirectoryServices.ActiveDirectoryAccessRule] $ace 
 
    #  walk DACL 
    foreach ($ace in $acl) 
    { 
        # trustee equals EnterpriseKeyAdmins group? 
        if ($ace.IdentityReference.Value -eq $Global:EnterpriseKeyAdmins) 
        {             
            Write-Host "`r`nFound matching AccessControlEntry:" 
 
            # write current EnterpriseKeyAdmins group AccessControlEntry to console 
            ReportAce $ace 
 
            # does current AccessControlEntry not equals desired AccessControlEntry? 
            [bool]$tempcheck = CheckAce $ace 
             
            if ($tempcheck -eq $true) 
            { $mustedit = $true } 
             
            Write-host "mustedit: $tempcheck" 
 
            if ($tempcheck -eq $true) 
            { 
                Write-Host "`r`nRemoving non-desired AccessControlEntry" 
                 
                # remove non-desired ACE from SD 
                $waste = $retsd.RemoveAccessRuleSpecific($ace) 
            } 
 
            else 
            { 
                # we found desired ACE 
                $found = $true 
            } 
        }        
    } 
 
    # we didn't find desired ACE 
    if ($found -ne $true) 
    {    
        # consider -Force switch 
        if ($mustedit -ne $true) 
        { $mustedit = $force } 
 
        # do we need to add the default ACE? 
        if ($mustedit -eq $true) 
        { 
            Write-Host "`r`nAdding desired AccessControlEntry:" 
                 
            # get default EnterpriseKeyAdmins group ACE 
            [System.DirectoryServices.ActiveDirectoryAccessRule] $defaultace =  BuildDefaultAce 
 
            # write default EnterpriseKeyAdmins group ACE to console 
            ReportAce $defaultace 
 
            # add default EnterpriseKeyAdmins group ACE 
            $retsd.AddAccessRule($defaultace) 
        } 
    } 
 
    # return must commit changes to AD 
    $ret.MustCommit = $mustedit 
 
    # return edited SD 
    $ret.NewSD = $retsd 
 
    return $ret 
} 
 
#region does current AccessControlEntry not equals desired AccessControlEntry? 
#endregion 
function CheckAce([System.DirectoryServices.ActiveDirectoryAccessRule] $ace) 
{ 
    [bool] $ret = $true 
 
    # get default EnterpriseKeyAdmins group ACE for compare  
    [System.DirectoryServices.ActiveDirectoryAccessRule] $ctrl = BuildDefaultAce 
 
    # compare AceType 
    if ($ace.AccessControlType -eq $ctrl.AccessControlType) 
    { 
        # compare AccessMask 
        if ($ace.ActiveDirectoryRights -eq $ctrl.ActiveDirectoryRights) 
        { 
            # compare ObjectFlags 
            if ($ace.ObjectFlags -eq $ctrl.ObjectFlags) 
            {                 
                # compare GUID of targted ObjectType 
                if ($ace.ObjectType.ToString() -eq $ctrl.ObjectType.ToString()) 
                { 
                    # compare InheritanceFlags 
                    if ($ace.InheritanceFlags -eq $ctrl.InheritanceFlags) 
                    { 
                        # compare InheritanceType 
                        $ret = ($ace.InheritanceType -eq $ctrl.InheritanceType) 
                    } 
 
                    else 
                    { 
                        $ret = $false 
                    } 
                } 
 
                else 
                { 
                    $ret = $false 
                } 
            } 
 
            else 
            { 
                $ret = $false 
            } 
        } 
 
        else 
        { 
            $ret = $false 
        } 
    } 
 
    else 
    { 
        $ret = $false 
    } 
 
 
    # return not match 
    return !$ret 
} 
 
#region get default EnterpriseKeyAdmins group ACE 
#endregion 
function BuildDefaultAce() 
{ 
    [System.DirectoryServices.ActiveDirectoryAccessRule] $ret = $null 
 
    [System.Security.Principal.SecurityIdentifier] $trustee = [System.Security.Principal.SecurityIdentifier]$Global:EnterpriseKeyAdmins 
 
    $ret = New-Object System.DirectoryServices.ActiveDirectoryAccessRule($trustee,  
                                                                         $Global:DesiredAccessMask,  
                                                                         $Global:DesiredAceType,  
                                                                         $Global:ObjectType,  
                                                                         $Global:DesiredInheritanceType) 
 
    return $ret 
} 
 
#region get FullControl EnterpriseKeyAdmins group ACE 
# !!!!! for testing purposes only !!!!! 
#endregion 
function BuildFullAce() 
{ 
    [System.DirectoryServices.ActiveDirectoryAccessRule] $ret = $null 
 
    [System.Security.Principal.SecurityIdentifier] $trustee = [System.Security.Principal.SecurityIdentifier]$Global:EnterpriseKeyAdmins 
 
    [System.Security.AccessControl.AccessControlType] $allow = [System.Security.AccessControl.AccessControlType]::Allow 
 
    [System.DirectoryServices.ActiveDirectoryRights] $fullcontrol = [System.DirectoryServices.ActiveDirectoryRights]::GenericAll 
 
    [System.DirectoryServices.ActiveDirectorySecurityInheritance] $inherit = [System.DirectoryServices.ActiveDirectorySecurityInheritance]::All 
 
    $ret = New-Object System.DirectoryServices.ActiveDirectoryAccessRule($trustee,  
                                                                         $fullcontrol,  
                                                                         $allow,   
                                                                         $inherit) 
 
    return $ret 
 
} 
 
#region  write ACE to console 
#endregion 
function ReportAce([System.DirectoryServices.ActiveDirectoryAccessRule] $ace) 
{ 
    Write-Host "`tTrustee: $($ace.IdentityReference)" 
 
    Write-Host "`tAceType: $($ace.AccessControlType)" 
 
    Write-Host "`tAcessMask: $($ace.ActiveDirectoryRights)" 
 
    Write-Host "`tInheritanceFlags: $($ace.InheritanceFlags)" 
 
    Write-Host "`tInheritanceType: $($ace.InheritanceType)" 
 
    Write-Host "`tObjectFlags: $($ace.ObjectFlags)" 
 
    Write-Host "`tObjectType: $($ace.ObjectType)" 
 
    Write-Host "`tInheritedObjectType: $($ace.InheritedObjectType)" 
 
    Write-Host "`tObjectFlags: $($ace.ObjectFlags)" 
} 
 
#endregion 
 
#region Globals 
 
# RID of the Enterprise Key Admnins group 
[string] $Global:EnterpriseKeyAdminsRid = 527 
 
# AceType:Allow 
[System.Security.AccessControl.AccessControlType] $Global:DesiredAceType = [System.Security.AccessControl.AccessControlType]::Allow 
 
# AccessMask ReadProperty | WriteProperty 
[System.DirectoryServices.ActiveDirectoryRights] $Global:DesiredAccessMask = [System.DirectoryServices.ActiveDirectoryRights]::ReadProperty -bor [System.DirectoryServices.ActiveDirectoryRights]::WriteProperty 
 
# Guid of msDS-KeyCredentialLink attribute 
[Guid] $Global:ObjectType = [Guid]"5b47d60f-6090-40b2-9f37-2a4de88f3063" 
 
# InheritanceType:All 
[System.DirectoryServices.ActiveDirectorySecurityInheritance] $Global:DesiredInheritanceType = [System.DirectoryServices.ActiveDirectorySecurityInheritance]::All 
 
# the constructed SID of Enterprise Key Admnins group 
[string] $Global:EnterpriseKeyAdmins = [String]::Empty 
 
#endregion 
 
 
#region Main 
 
$forestinfo = GetForest $ForestName 
 
if ($forestinfo.Success -eq $true) 
{     
    "Inspecting forest $($forestinfo.ForestName)" 
 
    $Global:EnterpriseKeyAdmins = "$($forestinfo.RootSid)-$($Global:EnterpriseKeyAdminsRid)" 
 
 
    foreach ($domain in $forestinfo.Domains) 
    { 
        HandleDomain $domain $Force 
    } 
} 
 
else 
{ 
    "ERROR: No domains found in $($ret.ForestName)!" 
} 
 
#endregion